#include <QtCore/QDebug>
#include <QtCore/QtPlugin>
#include <QtCore/QDateTime>
#include <QtWidgets/QWidget>
#include <QtWidgets/QTableWidget>
#include <QtWidgets/QTreeWidget>


// Кастомный non-UI тип
class BlueObject : public QObject
{
	Q_OBJECT

	// X.HMI UUID типа из юнита в плагине
	Q_CLASSINFO(QTNATIVE_ATTRIBUTE_UUID, "FB9855C8-246E-4176-9A29-557ABA1A0FED")


	Q_PROPERTY(bool angry READ angry WRITE setAngry NOTIFY becomeAngry)

public:

	Q_INVOKABLE BlueObject() : _angry()
	{
		qDebug() << "BlueObject()";
	}

	~BlueObject()
	{
		qDebug() << "~BlueObject()";
	}

	bool angry() const
	{
		qDebug() << "BlueObject::angry() returns " << _angry;
		return _angry;
	}

	void setAngry(bool angry)
	{
		qDebug() << "BlueObject::setAngry() with " << angry;

		if (angry == _angry)
			return;

		_angry = angry;

		becomeAngry(_angry);
	}

	Q_INVOKABLE QString dummyMethod(bool super, int, QString complicatedName)
	{
		return (super ? "TRUE" : "FALSE") + ("_" + complicatedName);
	}

	Q_INVOKABLE void methodWithAlias_1(uint8_t val)
	{
	}

	Q_INVOKABLE void methodWithAlias_2(::uint8_t val)
	{
	}

	Q_INVOKABLE void methodWithAlias_3(std::uint8_t val)
	{
	}

	Q_INVOKABLE void methodWithAlias_4(::std::uint8_t val)
	{
	}

	Q_INVOKABLE void disconnectAllSignals()
	{
		disconnect();
	}


Q_SIGNALS:

	void becomeAngry(bool);

private:

	bool _angry;

};

// Кастомный UI тип
class RedWidget : public QTableWidget
{
	Q_OBJECT

	// X.HMI UUID типа из юнита в плагине
	Q_CLASSINFO(QTNATIVE_ATTRIBUTE_UUID, "931EBDEF-3656-47B5-8390-4931B81FE17B")

	Q_PROPERTY(bool angry READ angry WRITE setAngry NOTIFY becomeAngry)

	Q_PROPERTY(QString fakeName READ fakeName WRITE setFakeName)
	Q_PROPERTY(QString fullFakeName READ fullFakeName)

public:

	bool angry() const
	{
		qDebug() << "RedWidget::angry() returns " << _angry;
		return _angry;
	}

	void setAngry(bool angry)
	{
		qDebug() << "RedWidget::setAngry() with " << angry;

		if (angry == _angry)
			return;

		_angry = angry;

		if (_angry)
			setStyleSheet("background-color:red;");
		else
			setStyleSheet("background-color:green;");

		becomeAngry(_angry);
	}

	QString fullFakeName() const
	{
		qDebug() << "RedWidget::fullFakeName() returns " << ("full" + _fakeName);
		return "full" + _fakeName;
	}

	QString fakeName() const
	{
		qDebug() << "RedWidget::fakeName() returns " << _fakeName;
		return _fakeName;
	}

	void setFakeName(QString const &fakeName)
	{
		qDebug() << "RedWidget::setFakeName() with " << fakeName;
		_fakeName = fakeName;
	}

public:

	Q_INVOKABLE RedWidget(QWidget *parent = nullptr)
		: QTableWidget(parent)
		, _angry(false)
	{

		qDebug() << "RedWidget(...)";
		setAngry(true);

		setRowCount(3);
		setColumnCount(2);

		setItem(0, 0, new QTableWidgetItem("Cock"));
		setItem(0, 1, new QTableWidgetItem("Wolf"));
		setItem(1, 0, new QTableWidgetItem("Monkey"));
		setItem(1, 1, new QTableWidgetItem("Otter"));
		setItem(2, 0, new QTableWidgetItem("Tortoise"));
		setItem(2, 1, new QTableWidgetItem("Bear"));
	}

	~RedWidget()
	{
		qDebug() << "~RedWidget()";
	}

	// Обычные методы (не слоты и не сигналы) нужно явно подмечать атрибутом `Q_INVOKABLE`
	// для того чтобы методы были доступны через мета-объектную систему. Конструкторов это тоже касается
	Q_INVOKABLE void setHappy(bool happy)
	{
		setAngry(!happy);

		dummySignal(_angry ? "SUPERANGRY" : "SUPERHAPPY");
	}

	Q_INVOKABLE bool getHappy() const
	{
		return !_angry;
	}

	Q_INVOKABLE QString dummyMethod(bool super, int, QString complicatedName)
	{
		return (super ? "true" : "false") + ("_" + complicatedName);
	}

	Q_INVOKABLE void disconnectAllSignals()
	{
		disconnect();
	}

Q_SIGNALS:

	void becomeAngry(bool);

	void dummySignal(QString const &);


public Q_SLOTS:

	void dummySlot(QString const &)
	{
	}

private:

	bool _angry;

	QString _fakeName;

};

// Кастомный UI-тип
class BlackWidget : public QTreeWidget
{
	Q_OBJECT

	Q_CLASSINFO(QTNATIVE_ATTRIBUTE_UUID, "08CE46AB-9CA3-450F-81D0-5FBCCFD95399")

	Q_PROPERTY(bool angry READ angry WRITE setAngry NOTIFY becomeAngry)

	Q_PROPERTY(QColor color READ getColor WRITE setColor)

	Q_PROPERTY(QDateTime dateTime READ getDateTime WRITE setDateTime)

public:

	Q_INVOKABLE BlackWidget(QWidget *parent = nullptr)
		: QTreeWidget(parent)
	{
		qDebug() << "BlackWidget(...)";

		headerItem()->setText(0, QDateTime::currentDateTimeUtc().toString(dtFormat));

		headerItem()->setText(1, "Some Header");

		setColumnCount(2);

		invisibleRootItem()->addChild(new QTreeWidgetItem(QStringList() << "Big" << "Elephant"));
		invisibleRootItem()->addChild(new QTreeWidgetItem(QStringList() << "Small" << "Mouse"));
	}

	~BlackWidget()
	{
		qDebug() << "~BlackWidget()";
	}

	bool angry() const
	{
		qDebug() << "BlackWidget::angry() returns " << _angry;
		return _angry;
	}

	void setAngry(bool angry)
	{
		qDebug() << "BlackWidget::setAngry() with " << angry;

		if (angry == _angry)
			return;

		_angry = angry;

		becomeAngry(_angry);
	}

	Q_INVOKABLE QString dummyMethod(bool super, QDateTime, int const &val = 5, QString complicatedName = QString())
	{
		return (super ? "True" : "False") + ("_" + complicatedName);
	}

	Q_INVOKABLE QColor getColor() const
	{
		qDebug() << "BlackWidget::getColor()";

		return headerItem()->foreground(0).color();
	}

	Q_INVOKABLE void setColor(QColor const &color)
	{
		qDebug() << "BlackWidget::setColor()";

		headerItem()->setForeground(0, color);
	}

	Q_INVOKABLE QDateTime getDateTime() const
	{
		qDebug() << "BlackWidget::getDateTime()";

		return QDateTime::fromString(headerItem()->text(0), dtFormat);
	}

	Q_INVOKABLE void setDateTime(QDateTime const &dt)
	{
		if (dt.isValid())
			return;

		qDebug() << "BlackWidget::setDateTime()";

		headerItem()->setText(0, dt.toString(dtFormat));
	}

	Q_INVOKABLE void refreshDateTime()
	{
		qDebug() << "BlackWidget::refreshDateTime()";

		headerItem()->setText(0, QDateTime::currentDateTimeUtc().toString(dtFormat));
	}


Q_SIGNALS:

	void becomeAngry(bool);

private:

	static auto constexpr dtFormat = "yyyy-MM-dd HH:mm:ss";

	bool _angry;

};
// Данный тип - точка входа в Qt плагин.
// Какой литерал после IID добавлять пользователь пусть уж сам решает, главное чтобы макрос был и этого
// достаточно, хостовое qt-приложение теперь может загрузить плагин через `QPluginLoader::instance()`
class FirstPlugin : public QObject
{
	Q_OBJECT

	// X.HMI UUID юнита из юнит-линка в плагине
	Q_CLASSINFO(QTNATIVE_ATTRIBUTE_UUID, "BF7BA334-E086-420B-8177-EF35B50E0506")

	Q_PLUGIN_METADATA(IID "org.qt-project.Qt.Examples.FirstPlugin")

public:

	// Экспортируемый метод для регистрации и передачи метатипов, через метаобъектную систему X.HMI ищет этот метод по сигнатуре
	Q_INVOKABLE QVariant QTNATIVE_REGISTER_TYPES()
	{
		// Чтобы полноценно работать с QObject подтипами в мета-объектной системе, достаточно одного лишь вызова `qRegisterMetaType<T*>()`.
		QVector<int> types;

		types.push_back(qRegisterMetaType<BlueObject *>());
		types.push_back(qRegisterMetaType<RedWidget *>());
		types.push_back(qRegisterMetaType<BlackWidget *>());

		qDebug() << "FirstPlugin registers metatypes " << types;

		return QVariant::fromValue(types);
	}

	// Конструкция представляющего Qt-плагин объекта может представлять собой инициализацию плагина
	FirstPlugin()
	{
		qDebug() << "FirstPlugin()";
	}

	~FirstPlugin()
	{
		qDebug() << "~FirstPlugin()";
	}
};

#include "firstplugin.moc"